#!/usr/bin/env python
#-*- coding:utf-8 -*-

import sys
import random
from libnum import len_in_bits, ceil
from ranges import Ranges

class BlocktypeAttack(object):
    def __init__(self, rsa_n, rsa_e, oracle, debug=True):
        self.n = rsa_n
        self.e = rsa_e
        self.oracle_func = oracle
        self.key_size = len_in_bits(self.n) / 8
        self.blocktype = 1 << (self.key_size * 8 - 16)
        self.window = 0, self.blocktype * 256 - 1

        self.orig_c = None
        self.c = None
        self.k = None
        self.blinding_s = None

        self.do_debug = debug
        self.oracle_requests_count = 0
        #self.responses = {s: k}

    def debug(self, *msg):
        if self.do_debug:
            sys.stderr.write(" ".join(map(str, msg)))
            sys.stderr.flush()

    def oracle(self, s):
        self.last_c = (self.c * pow(s, self.e, self.n)) % self.n
        self.oracle_requests_count += 1
        return self.oracle_func(self.last_c)

    def decrypt(self, rsa_c):
        self.orig_c = rsa_c % self.n
        self.c = self.orig_c

        self.debug("[*] Checking if the message is padded already...\n")
        self.k = self.oracle(1)
        if self.k == False:
            self.blind()
        self.debug("[*] Start message blocktype:", self.k, "\n")

        return self.reduce_iterative()

    def blind(self):
        self.debug("[*] Not padded, start blinding: ")
        k = False
        while k == False:
            s = random.randint(2, self.n - 1)
            k = self.oracle(s)
            self.debug(".")
        self.blinding_s = s
        self.c = self.last_c
        self.k = int(k)
        self.debug("\n")

    def search_for_s(self, start_from, upper_limit=None):
        s = start_from
        self.debug("[*] Searching for valid padding (starting from %s):\n" % s)
        while True:
            k = self.oracle(s)
            self.debug(".")
            if k != False:
                break
            s += 1

            if upper_limit is not None and s > upper_limit:
                self.debug("\n")
                return None

        self.debug("\n[*] Done: s = %s, k = %d (0x%02x)\n" % (s, k, k))
        return s, k

    def reduce_by_shift(self, s, k):
        shift = 8 - len_in_bits(k)
        off_bit = self.blocktype
        c = k * off_bit
        d = c + off_bit - 1

        if shift != 0:
            s = s * (1 << shift)
            k = self.oracle(s)
            if k == False:
                raise ValueError("Error occured while shifting! Bad oracle?")

            off_bit >>= shift
            c = k * off_bit
            d = c + off_bit - 1
            self.debug("[*] Guessed %d bits by shifting\n" % shift)
        return c, d

    def reduce_iterative(self):
        n = self.n
        BLOCK = self.blocktype

        a, b = self.reduce_by_shift(1, self.k)
        R = Ranges((a, b))
        
        s_prev = 1
        while True:
            if len(R.segments) == 1 and s_prev > 1: #TODO: parallel search in each interval
                self.debug("[*] One interval left, starting binary search\n")
                s, k = self.binary_reduce(R, s_prev)
            else:
                s, k = self.simple_reduce(R, s_prev)

            # *s is map from (a, b) to (c, d)
            c, d = self.reduce_by_shift(s, k)

            new_R = Ranges()
            for a, b in R.segments:
                r_min = ceil(a * s - d, n)
                r_max = (b * s - c) / n
                r_count = max(0, 1 + r_max - r_min)

                r = r_min
                while r <= r_max:
                    new_a = ceil(c + r * n, s)
                    new_b = (d + r * n) / s

                    new_R.add_range(new_a, new_b)

                    r += 1

                self.debug("[*] Added %d new intervals\n" % r_count)

            self.debug("[*] Intersecting...\n")
            R &= new_R

            size = R.len
            self.debug("[*] Total size: ...%s (%s bits) in %s segments\n" % (str(size)[-32:], len_in_bits(size), len(R.segments)))

            if size == 1:
                return R.min

            if len(R.segments) == 0:
                raise ValueError("Error occured! Bad oracle?")

            s_prev = s

    def simple_reduce(self, R, s_prev):
        # the smallest possible s is map
        #   from maximal possible M (range upper bound)
        #   to N + minimal detectable bound
        s_start = max(s_prev + 1, ceil(self.n + self.window[0], R.max))
        return self.search_for_s(s_start)

    def binary_reduce(self, R, s_prev):
        n = self.n

        a, b = R.min, R.max
        c, d = self.window

        r = ceil((b * s_prev - c) * 2, n)
        while True:
            s = ceil((c + r * n), b)
            end = (d + r * n) / a
            res = self.search_for_s(s, end)
            if res:
                return res
            r += 1